@@ -41,7 +41,8 @@ sub new {
             header_table => [],
 
             # HPACK. Header Table size
-            ht_size => 0,
+            ht_size     => 0,
+            max_ht_size => DEFAULT_HEADER_TABLE_SIZE,
 
             settings => {%default_settings},
 
@@ -53,7 +54,8 @@ sub new {
             header_table => [],
 
             # HPACK. Header Table size
-            ht_size => 0,
+            ht_size     => 0,
+            max_ht_size => DEFAULT_HEADER_TABLE_SIZE,
 
             # HPACK. Emitted headers
             emitted_headers => [],
@@ -101,6 +103,10 @@ sub new {
         }
     }
 
+    # Sync decode context max_ht_size
+    $self->{decode_ctx}->{max_ht_size} =
+      $self->{decode_ctx}->{settings}->{&SETTINGS_HEADER_TABLE_SIZE};
+
     $self;
 }
 
@@ -139,9 +145,9 @@ sub enqueue_first {
     my $self = shift;
     my $i    = 0;
     for ( 0 .. $#{ $self->{queue} } ) {
-        last
-          if ( ( $self->frame_header_decode( \$self->{queue}->[$_], 0 ) )[1] !=
-            CONTINUATION );
+        my $type =
+          ( $self->frame_header_decode( \$self->{queue}->[$_], 0 ) )[1];
+        last if $type != CONTINUATION && $type != PING;
         $i++;
     }
     while ( my ( $type, $flags, $stream_id, $data_ref ) = splice( @_, 0, 4 ) ) {
@@ -237,6 +243,12 @@ sub state_machine {
         }
     }
 
+    # Unexpected CONTINUATION frame
+    elsif ( $type == CONTINUATION ) {
+        tracer->error("Unexpected CONTINUATION frame\n");
+        $self->error(PROTOCOL_ERROR);
+    }
+
     # State machine
     # IDLE
     elsif ( $prev_state == IDLE ) {
@@ -275,6 +287,13 @@ sub state_machine {
         elsif ( $type == RST_STREAM ) {
             $self->stream_state( $stream_id, CLOSED );
         }
+        elsif ($type == HEADERS
+            && !$pending
+            && $self->stream_trailer($stream_id) )
+        {
+            tracer->error("expected END_STREAM flag for trailer HEADERS frame");
+            $self->error(PROTOCOL_ERROR);
+        }
     }
 
     # RESERVED (local/remote)
@@ -478,4 +497,16 @@ sub ack_ping {
     $self->enqueue_first( PING, ACK, 0, $payload_ref );
 }
 
+sub send_ping {
+    my ( $self, $payload ) = @_;
+    if ( !defined $payload ) {
+        $payload = pack "C*", map { rand(256) } 1 .. PING_PAYLOAD_SIZE;
+    }
+    elsif ( length($payload) != PING_PAYLOAD_SIZE ) {
+        $payload = sprintf "%*.*s",
+          -PING_PAYLOAD_SIZE(), PING_PAYLOAD_SIZE, $payload;
+    }
+    $self->enqueue( PING, 0, 0, \$payload );
+}
+
 1;
