@@ -548,4 +548,23 @@ sub feed {
     substr( $self->{input}, 0, $offset ) = '' if $offset;
 }
 
+=head3 ping
+
+Send ping frame to client (to keep connection alive)
+
+    $server->ping
+
+or
+
+    $server->ping($payload);
+
+Payload can be arbitrary binary string and must contain 8 octets. If payload argument
+is omitted server will send random data.
+
+=cut
+
+sub ping {
+    shift->{con}->send_ping(@_);
+}
+
 1;
